=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Designed February 2015 by Fredo6

# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   body_Lib6Solid.rb
# Original Date	:   07 Feb 15
# Description	:   Manage solids and pseudo-solids (implementation)
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module Traductor

T6[:ERR_DuringAutoReverse] = "Error while processing Auto-Reverse"
T6[:TIT_AutoReverse] = "Auto-Reverse Faces"

T6[:ERR_DuringConvexify] = "Error while processing Convexify"
T6[:TIT_Convexify] = "Convexify 3D Shapes"

T6[:MSG_SolidProcessSelection] = "Processing Selection"
T6[:MSG_SolidCalculateGroupings] = "Calculating Groupings:"
T6[:MSG_SolidCreateConvexifiers] = "Creating Convexify robots"
T6[:MSG_SolidConvexify] = "Convexify 3D Shapes"
T6[:MSG_SolidAutoReverse] = "Auto-Reverse Shapes"

#=============================================================================================
#=============================================================================================
# Class AutoReverseFaces: main class for Auto Reverse of faces orientation: on selection
#=============================================================================================
#=============================================================================================

class AutoReverseFaces

#INIT: Class initialization
def initialize__(*hargs)
	@model = Sketchup.active_model
	
	#Parsing the arguments
	hargs.each { |harg| harg.each { |key, value| parse_args(key, value) } if harg.class == Hash }
	
	#Text initialization
	init_messages
	
	#Creating the auto_reverser
	@auto_reverser = AutoReverseFacesCompact.new
	
	#Creating the Progress bar tool
	pgtool_notify_proc = self.method "notify_event"
	hsh = { :interruptible => true, :notify_proc => pgtool_notify_proc }
	@pgtool = Traductor::ProgressBarTool.new hsh
	
	@model.tools.push_tool @pgtool	
	
	#Creating an Operation framework if not provided
	unless @suops
		@suops_local = true
		hsh = { :title => @menutitle, :no_commit => true, :end_proc => self.method('robot_terminate') }
		@suops = Traductor::SUOperation.new hsh
	end	

end

#INIT: Initializing messages
def init_messages
	@menutitle = T6[:TIT_AutoReverse]
	@msg_selection = T6[:MSG_SolidProcessSelection]
	@msg_grouping = T6[:MSG_SolidCalculateGroupings]
	@msg_autoreverse = T6[:MSG_SolidAutoReverse]
end

#INIT: Parse the arguments of the initialize method
def parse_args(key, value)
	skey = key.to_s
	case skey
	when /suops/i
		@suops = value
	when /notify_exit_proc/i
		@notify_exit_proc = value
	end	
end

#INIT: Top method to process the Convexifying of Selection
def auto_reverse_selection(selection=nil)	
	@model = Sketchup.active_model
	@selection = @model.selection
	@view = @model.active_view
	@tr_id = Geom::Transformation.new

	#Starting Robot
	@suops.start_execution { robot_execute }
end
	
#---------------------------------------------------------------------------------------------
# ROBOT_INSPECTION: Robot for Inspection
#---------------------------------------------------------------------------------------------

#ROBOT: Main State-robot function to progress along the execution
def robot_execute
	begin
		robot_execute_protected
	rescue Exception => e
		Traductor::RubyErrorDialog.invoke e, @menutitle, T6[:ERR_DuringAutoReverse]
		notify_event :abort
	end
end

#ROBOT: Main State-robot function to progress along the inspection (protected)
def robot_execute_protected
	while(action, *param = @suops.current_step) != nil	
		case action
		when :_init
			@nreversed = 0
			next_step = :process_selection
			
		#Exploring the selection
		when :process_selection
			return if @suops.yield?
			selection = (@selection.empty?) ? @model.active_entities : @selection.to_a
			@lst_comp_info = []
			@hsh_cdef = {}
			notify_event action
			process_selection selection, nil, @tr_id
			next_step = :grouping
		
		#Compute_groupings
		when :grouping
			return if @suops.yield?
			@lst_groupings = []
			@lst_comp_info.each do |faces, comp, tr|
				@lst_groupings += calculate_groupings(faces, comp, tr)
				notify_event action, @lst_groupings.length
			end
			next_step = [:auto_reverse, 0]
	
		#Creating the list of ConvexifierCompacts
		when :auto_reverse
			return if @suops.yield?
			igrouping, = param
			faces, = @lst_groupings[igrouping]
			if faces
				notify_event action, igrouping
				@nreversed += @auto_reverser.process_faces faces
				next_step = [:auto_reverse, igrouping+1]
			else
				next_step = :finished
			end	
				
		#End of Processing
		when :finished
			notify_event :exit
			next_step = nil
		end	
		
		break if @suops.next_step(*next_step)
	end
end
	
#ROBOT: Termination routine	
def robot_terminate(time)
	if time
		notify_event :exit
	else
		notify_event :abort
	end
end	
	
#ROBOT: Event notifier, displaying status in the Progression panel
def notify_event(event, *args)		
	case event
	
	#Normal Exit
	when :exit
		@pgtool.exit
		@notify_exit_proc.call :exit if @notify_exit_proc
		(@nreversed == 0) ? @suops.abort_operation : @suops.commit_operation
		
	#Abort due to error or interruption by user	
	when :abort
		@pgtool.exit
		@notify_exit_proc.call :abort if @notify_exit_proc
		@suops.abort_operation
		
	#Process Selection	
	when :process_selection
		@pgtool.panel_progression @msg_selection, @menutitle

	#Process Groupings	
	when :grouping
		nb_grouping, = args
		@pgtool.panel_progression "#{@msg_grouping} #{nb_grouping}"
	
	#Processing AutoReverse of compact solid	
	when :auto_reverse
		igrouping, = args
		message = "#{@msg_autoreverse} #{igrouping} / #{@lst_groupings.length}"
		@pgtool.panel_progression message

	when :button_up, :cancel, :undo, :key_up
		@suops.interrupt?
		
	end	
end
	
#---------------------------------------------------------------------------------------------
# ALGO: Processing algorithm
#---------------------------------------------------------------------------------------------
	
#ALGO: For each component / group, split the faces into connected groupings
def calculate_groupings(faces, comp, tr)
	return [] if faces.empty?
	lst_groups = []
	hfaces_used = {}
	
	hfaces = {}
	faces.each { |f| hfaces[f.entityID] = f }
	
	faces.each do |face|
		next if hfaces_used[face.entityID]
		con_faces = face.all_connected.grep(Sketchup::Face).find_all { |f| hfaces[f.entityID] }
		con_faces.each { |f| hfaces_used[f.entityID] = true }
		lst_groups.push [con_faces, comp, tr]
	end	
	lst_groups
end

#ALGO: Recursive exploration	
def process_selection(entities, comp, tr)
	faces = entities.grep(Sketchup::Face)
	@lst_comp_info.push [faces, comp, tr] unless faces.empty?
	
	entities.each do |e|
		if e.instance_of?(Sketchup::Group)
			gent = G6.grouponent_make_unique(e)
			process_selection gent, e, tr * e.transformation
		elsif e.instance_of?(Sketchup::ComponentInstance)
			cdef = e.definition
			next if @hsh_cdef[cdef.entityID]
			process_selection cdef.entities, e, tr * e.transformation
			@hsh_cdef[cdef.entityID] = true 
		end			
	end
end
	
end	#class AutoReverseFaces

#=============================================================================================
#=============================================================================================
# Class AutoReverseFaceCompact: main class for Auto Reverse: on compact solids
#=============================================================================================
#=============================================================================================

class AutoReverseFacesCompact

#INIT: Class initialization
def initialize__(*hargs)
	@tr_id = Geom::Transformation.new

end

#Top method to auto-reverse a compact set of faces
def process_faces(lst_faces, tr=nil)
	nreversed = 0
	@tr = (tr) ? tr : @tr_id
	
	#Classifying the faces
	lst_good, lst_rev_good, lst_zero, lst_others = classify_faces(lst_faces)
	
	#Reversing the faces that should be ok if reversed
	lst_rev_good.each { |f| G6.reverse_face_with_uv(f) }
	nreversed += lst_rev_good.length
	
	#Checking the list of new good faces
	lst_new_good = lst_good + lst_rev_good
	if lst_new_good.empty?
		face0 = lst_zero.find { |f| G6.front_face_is_visible?(f, @tr) }
		if face0
			lst_zero.delete face0
		else	
			face0 = lst_others.find { |f| G6.front_face_is_visible?(f, @tr) }
			return nreversed unless face0
			lst_others.delete face0
		end	
		lst_new_good = [face0]
	end			
	
	#lst_new_good.each { |f| f.material = 'lightgreen' }
	
	#Propagate the face normal direction from the good faces
	nreversed += propagate_face_directions(lst_zero + lst_others, lst_new_good)

	nreversed
end

#Classify faces based on the the number of intersections
def classify_faces(lst_faces)
	lst_good = []
	lst_rev_good = []
	lst_zero = []
	lst_others = []
	lst_faces.each do |face|	
		nfront, nback = face_count_intersection(face, lst_faces)
		if nfront == 0 && nback.modulo(2) == 1
			lst_good.push face
		elsif nback == 0 && nfront.modulo(2) == 1
			lst_rev_good.push face
		elsif nfront == 0 && nback == 0
			lst_zero.push face
		else
			lst_others.push face
		end	
	end
	[lst_good, lst_rev_good, lst_zero, lst_others]
end

#Count intersection of the normal of a face with other faces in the front and back direction
def face_count_intersection(face0, lst_faces)
	center = face0.bounds.center
	center = face0.vertices[0].position unless [1, 2, 4].include?(face0.classify_point(center))
	normal0 = face0.normal
	line = [center, normal0]
	
	lpt_front = []
	lpt_back = []
	lst_faces.each do |face|
		next if face == face0
		ptinter = Geom.intersect_line_plane(line, face.plane)
		next unless ptinter
		ps = center.vector_to(ptinter).normalize % normal0
		next if ps == 0 || ![1, 2, 4].include?(face.classify_point(ptinter))
		if ps > 0
			lpt_front.push ptinter unless lpt_front.include?(ptinter)
		else	
			lpt_back.push ptinter unless lpt_front.include?(ptinter)
		end	
	end	
	[lpt_front.length, lpt_back.length]
end

#Expand face flipping from a list of good faces
def propagate_face_directions(lst_faces, good_faces)
	return 0 if lst_faces.empty?
	
	nreversed = 0
	hgood_faces = {}
	hless_good_faces = {}
	hfaces_used = {}
	good_faces.each { |f| hgood_faces[f.entityID] = true ; hfaces_used[f.entityID] = true }
	
	#Loop on faces
	#  - First pass if for real good faces
	#  - Second pass is for less good faces if any faces are left
	lfaces = lst_faces.clone
	for i in 0..1
		break if lfaces.empty?
		n = 0
		hgood_faces.update hless_good_faces if i == 1
		nmax = lfaces.length
		while lfaces.length > 0
			n += 1
			break if n > nmax
			face0 = lfaces.shift
			break unless face0
			face0_id = face0.entityID
			next if hfaces_used[face0_id]
			
			face0.edges.each do |e|
				#Finding a neighbour face which is good
				face2 = e.faces.find { |f| f != face0 && hgood_faces[f.entityID] }
				next unless face2
				
				#Checking if face0 must be reversed from edge / face orientation
				if e.reversed_in?(face0) == e.reversed_in?(face2)
					G6.reverse_face_with_uv(face0)
					nreversed += 1
				end	
				
				#If edge has only 2 faces, then face0 is now considered good. Otherwise suspicious 
				if e.faces.length == 2
					hgood_faces[face0_id] = true
				else
					hless_good_faces[face0_id] = true
				end
				
				#Face has been treated
				hfaces_used[face0_id] = true
				
				#Updating the limit for while loop
				n = 0
				nmax = lfaces.length
				break
			end
			lfaces.push face0 unless hfaces_used[face0_id]
		end	
	end	
		
	nreversed
end

end	#class AutoReverseFacesCompact

#=============================================================================================
#=============================================================================================
# Class SolidConvexify: Class managing the convexifying of a selection
#=============================================================================================
#=============================================================================================

class SolidConvexify

#Information on Grouping
GroupingInfo = Struct.new :id, :comp, :tr, :faces, :nb_faces, :lst_convex, :master_group, :parent

#Initialize and execute the marking of vertices
def initialize__(*hargs)	
	@model = Sketchup.active_model
	
	#Groups and layers
	@hsh_parents = {}
	@hsh_master_groups = {}
	
	#Parsing the arguments
	hargs.each { |harg| harg.each { |key, value| parse_args(key, value) } if harg.class == Hash }
	
	#Text initialization
	init_messages
	
	#Creating the Progress bar tool
	pgtool_notify_proc = self.method "notify_event"
	hsh = { :interruptible => true, :notify_proc => pgtool_notify_proc }
	@pgtool = Traductor::ProgressBarTool.new hsh
	@model.tools.push_tool @pgtool	
	
	#Creating an Operation framework if not provided
	unless @suops
		@suops_local = true
		hsh = { :title => @menutitle, :no_commit => true, :end_proc => self.method('robot_terminate') }
		@suops = Traductor::SUOperation.new hsh
	end	

end

#INIT: Initialize messages
def init_messages
	@menutitle = T6[:TIT_Convexify]
	@msg_selection = T6[:MSG_SolidProcessSelection]
	@msg_grouping = T6[:MSG_SolidCalculateGroupings]
	@msg_create_convexifiers = T6[:MSG_SolidCreateConvexifiers]
	@msg_convexify = T6[:MSG_SolidConvexify]
end

#INIT: Parse the arguments of the initialize method
def parse_args(key, value)
	skey = key.to_s
	case skey
	when /suops/i
		@suops = value
	when /notify_exit_proc/i
		@notify_exit_proc = value
	end	
end

#INIT: Top method to process the Convexifying of Selection
def convexify(selection=nil, *hparams)	
	@model = Sketchup.active_model
	@entities = @model.active_entities
	@selection = @model.selection
	@view = @model.active_view
	@tr_id = Geom::Transformation.new

	#Starting Robot
	@suops.start_execution { robot_execute @hparams, *hparams}
end
	
#---------------------------------------------------------------------------------------------
# INFO: Method for information
#---------------------------------------------------------------------------------------------

#INFO: return the current parameters used	
def get_parameters ; @hparams ; end
	
def grouping_info ; @lst_groupings ; end
	
def calculation_time ; @time_calc ; end
		
#---------------------------------------------------------------------------------------------
# ROBOT_INSPECTION: Robot for Inspection
#---------------------------------------------------------------------------------------------

#ROBOT: Main State-robot function to progress along the execution
def robot_execute(*hparams)
	@hparams = {}
	hparams.each { |hsh| @hparams.update hsh if hsh.class == Hash }
	
	begin
		robot_execute_protected
	rescue Exception => e
		Traductor::RubyErrorDialog.invoke e, @menutitle, T6[:ERR_DuringConvexify]
		notify_event :abort
	end
end

#ROBOT: Main State-robot function to progress along the inspection (protected)
def robot_execute_protected
	while(action, *param = @suops.current_step) != nil	
		case action
		when :_init
			@status_cvx = 0
			next_step = :process_selection
			
		#Exploring the selection
		when :process_selection
			return if @suops.yield?
			selection = (@selection.empty?) ? @model.active_entities : @selection.to_a
			@lst_comp_info = []
			@hsh_cdef = {}
			notify_event action
			prepare_layer
			process_selection selection, nil, @tr_id
			next_step = :grouping
		
		#Compute_groupings
		when :grouping
			return if @suops.yield?
			@lst_groupings = []
			@lst_comp_info.each do |faces, comp, tr|
				calculate_groupings(faces, comp, tr)
				notify_event action, @lst_groupings.length
			end
			next_step = :create_convexifier
	
		#Creating the list of ConvexifierCompacts
		when :create_convexifier
			return if @suops.yield?
			notify_event action
			@lst_convexifiers = []
			@lst_groupings.each do |grouping|
				grouping.master_group = master_group = get_master_group(grouping.comp)
				@lst_convexifiers.push SolidConvexifyCompact.new(grouping.faces, master_group, grouping.tr, @hparams)
			end	
			next_step = [:compact, 0]
		
		#Convexifying compact solids
		when :compact
			return if @suops.yield?
			icvx, = param
			cvx = @lst_convexifiers[icvx]
			if cvx
				notify_event action, icvx
				status = cvx.solid_split_robot
				
				#Continue splitting
				if status
					notify_event :iteration, status
					@status_cvx += 1
					
				#Splitting finished. Go the next one	
				else
					grouping= @lst_groupings[icvx]
					grouping.lst_convex = cvx.get_convex_info
					grouping.parent = cvx.get_top_group
					icvx += 1
				end	
				next_step = [:compact, icvx]
			else
				next_step = :finished
			end	
		
		#End of Processing
		when :finished
			#notify_event :exit
			next_step = nil
		end	
		
		break if @suops.next_step(*next_step)
	end
end
	
#ROBOT: Termination routine	
def robot_terminate(time)
	@time_calc = time
	if time
		notify_event :exit
	else
		notify_event :abort
	end
end	
	
#ROBOT: Event notifier, displaying status in the Progression panel
def notify_event(event, *args)	
	
	case event
	
	#Normal Exit
	when :exit
		@layer.visible = false if @layer
		if @status_cvx == 0
			@pgtool.panel_abort
			@suops.abort_operation
		else
			@pgtool.panel_commit
			@suops.commit_operation
		end	
		@pgtool.exit
		@notify_exit_proc.call :exit if @notify_exit_proc
		
	#Abort due to error or interruption by user	
	when :abort
		@pgtool.exit
		@notify_exit_proc.call :abort if @notify_exit_proc
		@suops.abort_operation
		
	#Process Selection	
	when :process_selection
		@pgtool.panel_progression @msg_selection, @menutitle

	#Process Groupings	
	when :grouping
		nb_grouping, = args
		@pgtool.panel_progression "#{@msg_grouping} #{nb_grouping}"
	
	#Create Convexifier structures	
	when :create_convexifier
		@pgtool.panel_progression @msg_create_convexifiers
		
	#Processing convexify of Solid	
	when :compact
		icvx, = args
		message = "#{@msg_convexify} #{icvx+1} / #{@lst_convexifiers.length}"
		@pgtool.panel_progression message
		
	when :iteration	
		status, = args
		@pgtool.panel_update_time "Iteration = #{status}"
		
	when :button_up, :cancel, :undo, :key_up
		@suops.interrupt?
	end	
end
	
#---------------------------------------------------------------------------------------------
# ALGO: Processing algorithm
#---------------------------------------------------------------------------------------------
	
#ALGO: Create a Grouping information structure	
def grouping_info_create(faces, comp, tr)
	gpinfo = GroupingInfo.new
	gpinfo.id = @lst_groupings.length
	@lst_groupings.push gpinfo
	gpinfo.faces = faces
	gpinfo.nb_faces = faces.length
	gpinfo.comp = comp
	gpinfo.tr = tr
	gpinfo
end
		
#ALGO: For each component / group, split the faces into connected groupings
def calculate_groupings(faces, comp, tr)
	return [] if faces.empty?
	lst_groupings = []
	hfaces_used = {}
	
	hfaces = {}
	faces.each { |f| hfaces[f.entityID] = f }
	
	faces.each do |face|
		next if hfaces_used[face.entityID]
		con_faces = face.all_connected.grep(Sketchup::Face).find_all { |f| hfaces[f.entityID] }
		con_faces.each { |f| hfaces_used[f.entityID] = true }
		grouping_info_create(con_faces, comp, tr)
		#lst_groupings.push [con_faces, comp, tr]
	end	
	#lst_groupings
end

#ALGO: Recursive exploration	
def process_selection(entities, comp, tr)
	faces = entities.grep(Sketchup::Face)
	@lst_comp_info.push [faces, comp, tr] unless faces.empty?
	
	entities.each do |e|
		if e.instance_of?(Sketchup::Group)
			gent = G6.grouponent_make_unique(e)
			parent_register(e, comp)
			process_selection gent, e, tr * e.transformation
		elsif e.instance_of?(Sketchup::ComponentInstance)
			cdef = e.definition
			next if !@layer && @hsh_cdef[cdef.entityID]
			parent_register(e, comp)
			process_selection cdef.entities, e, tr * e.transformation
			@hsh_cdef[cdef.entityID] = true 
		end			
	end
end
	
def parent_register(comp, parent)
	@hsh_parents[comp.entityID] = parent
end

def parent_of(comp)
	(comp) ? @hsh_parents[comp.entityID] : nil
end

def get_master_group(comp)
	return comp unless @layer
	master_group = @hsh_master_groups[comp]
	return master_group if master_group
	
	if comp
		parent = parent_of(comp)
		mg = get_master_group(parent)
		gent = (mg) ? mg.entities : @entities
		master_group = gent.add_group
		master_group.transformation = comp.transformation
	else
		master_group = @entities.add_group
		master_group.layer = @layer if master_group
	end
	@hsh_master_groups[comp] = master_group
	master_group	
end

#ALGO: Preparation of the layer if any
def prepare_layer
	@use_layer = @hparams[:use_layer]
	return unless @use_layer
	@layer_name = @hparams[:layer_name]
	return unless @layer_name && !@layer_name.empty?
	
	layers = @model.layers
	@layer = layers[@layer_name]
	@layer = layers.add @layer_name unless @layer
	@layer.visible = true
	
	get_master_group(nil)
	
	@hparams[:keep_original] = true
	@hparams[:layer] = @layer
end
	
end	#class SolidConvexify
	
#=============================================================================================
#=============================================================================================
# Class SolidConvexifyCompact: Class managing a compact set of faces
#=============================================================================================
#=============================================================================================

class SolidConvexifyCompact

PseudoTriangles = Struct.new :id, :lipt, :lpt 

#Structure for Concave Edge information
ConcaveEdgeInfo = Struct.new :edge, :vec_dir, :vec_bissec, :vec_bissec_normal, :lptref, 
                             :angle, :angle_perp, :len, :normal1, :normal2, :plane1, :plane2, 
							 :iplane_max, :nplane_max, :plane_area, :face1, :face2,
                             :connected, :nb_connected, :best_plane, :used, :rank, :ls_normal
							 
#INIT: Class instance initialization
def initialize__(faces, comp, tr, *hargs)
	@model = Sketchup.active_model
	@view = @model.active_view
	@tr_id = Geom::Transformation.new
	@entities = G6.grouponent_entities comp
	@lst_faces = faces
	@lst_faces = @entities.grep(Sketchup::Face) unless faces
	@comp = comp
	@tr = tr
	
	#Parsing the arguments
	hargs.each { |harg| harg.each { |key, value| parse_params(key, value) } if harg.class == Hash }

	#Tolerance for concavity
	tol_concave = @param_tolerance_concave
	@angle_concave_max = Math::PI - tol_concave.degrees if tol_concave && tol_concave != 0
	
	#Autoreverser
	@auto_reverser = Traductor::AutoReverseFacesCompact.new
	
	#Attributes used by the algorithm
	@dico = "Fredo6_Convexify"
	@attr_face_side = "Face side"
	
	#Colors
	init_colors
end

#INIT: Initialize colors and material
def init_colors
	#Colors for Cut plane
	materials = @model.materials
	name = "matPlane"
	@mat_plane = materials[name]
	unless @mat_plane
		@mat_plane = materials.add name
		color = Sketchup::Color.new('red')
		color.alpha = 0.5
		@mat_plane.color = color
	end	
end

#INIT: Return the list of convex shapes created
def get_convex_info
	@lst_solid_convex
end	

def get_top_group
	@top_group
end
	
#INIT: Top level API to make the solid convex
def preparation
	@status_face = 0
	@status_solid = false
	lst_faces = @lst_faces
	@group_plane = nil
	@lst_solid_convex = []
					
	#Making a copy of the original solid
	@top_group = solid_copy_group(lst_faces)

	#Aligning the orientation of faces
	lst_faces = @top_group.entities.grep(Sketchup::Face)
	nreversed = @auto_reverser.process_faces lst_faces
	hsh_edges = {}
	lst_faces.each { |face| face.edges.each { |e| hsh_edges[e.entityID] = e } }
	
	#Building the concave information	
	concave_build_info(hsh_edges.values)
	
	#Erasing the original faces
	if !@param_keep_original
		lerase = []
		@lst_faces.each { |face| lerase += [face] + face.edges }
		@entities.erase_entities lerase
	end
	
	###Detecting all faces with are concave and making them convex (for the future)
	###lst_faces, status_face = face_convexify_all(lst_faces)
	
	#Creating intersections in the solid recursively
	@lst_groups = [@top_group]
	
	#Returning the result
	true
end

#INIT: Parse the arguments of the initialize method
def parse_params(key, value)
	skey = key.to_s
	case skey
	when /animation/i
		@param_animation = value if G6.su_capa_refresh_view
	when /keep_original/i
		@param_keep_original = value
	when /box_style/i
		@param_box_style = value
	when /tolerance_concave/
		@param_tolerance_concave = value
	when /layer_name/
		@layer_name = value
	when /use_layer/
		@use_layer = value
	when /\Alayer\Z/
		@layer = value
	end	
end

#--------------------------------------------------------------
# SOLID: Processing for splitting the solid
#--------------------------------------------------------------

#SOLID: top level method for splitting a whole solid in one pass
def make_convex
	#Split processing - all in one call
	n = 0
	while solid_split_robot 
		n += 1
	end
	(n == 0) ? nil : n
end

#SOLID: top level method for splitting a whole solid (one pass only)
def solid_split_robot
	#Preparing the splitting (done only once)
	unless @lst_solid_convex
		return nil unless preparation
		@iteration = 0
	end

	@iteration += 1
	#return nil if @iteration > 1
	
	#Getting the next group to split
	g = @lst_groups.shift
	return nil unless g
		
	#Splitting the group into convex parts
	split_groups = solid_split(g)
	
	#New groups are put in the current working list
	if split_groups
		@lst_groups.concat split_groups
		
	#Group was actually convex	
	elsif g.valid?
		@lst_solid_convex.push g
		return @iteration
	end	
	
	#Erasing the group used for animation
	if @param_animation
		@view.refresh
		sleep @param_animation
	end	
	
	#Erasing the group used for animation
	@group_plane.erase! if @group_plane
	@group_plane = nil
	
	#Returning the status
	@iteration
end

#SOLID: Computation of the split of the group via a cut plane
def solid_split(top_group)
	top_gent = top_group.entities
	lst_faces = top_gent.grep(Sketchup::Face)
	return nil if lst_faces.length <= 1
	
	#Computing the best cut plane from concave edges
	best_plane = concave_best_cut_plane(top_group)
	
	#No need to cut - Solid is convex
	return nil unless best_plane
	
	#Preparing the Cut group
	t0 = Time.now
	lptref, vec_plane = best_plane
	ptref1, ptref2 = lptref
	diagonal = top_group.bounds.diagonal
	cut_plane = [ptref1, vec_plane]
	info_big = [ptref1, vec_plane, top_group.bounds.diagonal]
	cut_group = solid_prepare_cut_plane(ptref1, vec_plane, diagonal)
	cut_gent = cut_group.entities
	
	#Intersecting the cut plane with the solid and putting the edges into the top group
	top_gent.intersect_with false, @tr_id, top_gent, @tr_id, false, cut_group
	
	#Erasing the cut plane group
	cut_group.erase!
	
	#Splitting the faces into 3 sets, on each side of the plane
	#  - hfaces1, hfaces2: all faces strictly on each side of the plane
	#  - lst_touching_faces1, lst_touching_faces2: subset of faces touching the cut plane
	#  - section_faces: section faces located exactly on the cut plane
	hfaces1 = {}
	hfaces2 = {}
	lst_touching_faces1 = []
	lst_touching_faces2 = []
	section_faces = []
	
	top_gent.grep(Sketchup::Face).each do |face|
		side, touch = solid_face_which_side_of_plane(face, cut_plane) 
		if side == 1
			hfaces1[face.entityID] = face
			lst_touching_faces1.push face if touch
		elsif side == 2
			hfaces2[face.entityID] = face
			lst_touching_faces2.push face if touch
		else
			section_faces.push face
		end	
	end	
	
	#Special cases - one of the side is empty: 
	#Create a group with all faces and the other just with the section faces
	lsg = solid_split_special(hfaces1.values, hfaces2.values, section_faces, top_group)
	return lsg if lsg != :not_special
	
	#Calculating the groupings for faces on each side. Only the main group is close to the concave edge
	hgf1_main, hgf1_others = solid_face_grouping(hfaces1, lst_touching_faces1, ptref1, ptref2)
	hgf2_main, hgf2_others = solid_face_grouping(hfaces2, lst_touching_faces2, ptref1, ptref2)
		
	#Marking the faces back to their original grouping
	hgf1_others.each { |hgf| hfaces2.update(hgf) ; hfaces1.delete_if { |fid, f| hgf[fid] } }
	hgf2_others.each { |hgf| hfaces1.update(hgf) ; hfaces2.delete_if { |fid, f| hgf[fid] } }
	
	#Marking the faces of each group
	hfaces1.values.each { |f| f.set_attribute(@dico, @attr_face_side, 1)  }
	hfaces2.values.each { |f| f.set_attribute(@dico, @attr_face_side, 2)  }
	
	#Recomputing the touch faces
	lst_touching_faces = lst_touching_faces1 + lst_touching_faces2
	lst_touching_faces1 = lst_touching_faces.find_all { |f| hgf1_main[f.entityID] }
	lst_touching_faces2 = lst_touching_faces.find_all { |f| hgf2_main[f.entityID] }
	
	#Removing coplanar_edges on cut plane for touching faces when within the same group
	hedges = {}
	lst_touching_faces1.each { |f| f.edges.each { |e| hedges[e.entityID] = e } }
	lst_touching_faces2.each { |f| f.edges.each { |e| hedges[e.entityID] = e } }
	
	lerase = []
	hedges.each do |eid, e|
		f1, f2 = e.faces
		next unless f2
		f1id = f1.entityID
		f2id = f2.entityID
		next unless (hfaces1[f1id] && hfaces1[f2id]) || (hfaces2[f1id] && hfaces2[f2id])
		lerase.push e if e.start.position.on_plane?(cut_plane) && e.start.position.on_plane?(cut_plane) && G6.edge_coplanar?(e)
	end
	top_gent.erase_entities lerase
	
	#Final faces in each group
	lfaces = top_gent.grep(Sketchup::Face)
	lst_faces1 = lfaces.find_all { |f| f.get_attribute(@dico, @attr_face_side) == 1  }
	lst_faces2 = lfaces.find_all { |f| f.get_attribute(@dico, @attr_face_side) == 2  }
	
	#Creating the Group1 and Group2 and Transferring the faces
	g1 = @entities.add_group
	gent1 = g1.entities
	solid_transfer_faces(lst_faces1, gent1)
	solid_manage_tiny_edges(gent1, cut_plane)
	solid_transfer_section_faces(gent1, cut_plane, vec_plane, false, diagonal)
	
	g2 = @entities.add_group
	gent2 = g2.entities
	solid_transfer_faces(lst_faces2, gent2)
	solid_manage_tiny_edges(gent2, cut_plane)
	solid_transfer_section_faces(gent2, cut_plane, vec_plane.reverse, true, diagonal)
	
	#One of the two group is empty
	if lst_faces1.empty? || lst_faces2.empty?
		g1.erase!
		g2.erase!
		return nil
	end
	
	#Erasing the original top group
	top_group.erase!
	
	#Returning the two groups after checking their disjointed faces
	lg = solid_disjointed_groups(g1) + solid_disjointed_groups(g2)
	lg
end

#UTIL: Special case. Nothing on one side of the group
def solid_split_special(lfaces1, lfaces2, section_faces, top_group)
	return :not_special if lfaces1.length > 0 && lfaces2.length > 0
	return nil if section_faces.empty?
	if lfaces1.empty?
		g2 = @entities.add_group
		solid_transfer_faces(lfaces2, g2.entities)
		g1 = @entities.add_group
		solid_transfer_faces(section_faces, g1.entities)
	else	
		g1 = @entities.add_group
		solid_transfer_faces(lfaces1, g1.entities)
		g2 = @entities.add_group
		solid_transfer_faces(section_faces, g2.entities)
	end	
	top_group.erase!
	return solid_disjointed_groups(g1) + solid_disjointed_groups(g2)
end

#--------------------------------------------------------------
# UTIL: Utility methods used by the main algorithm
#--------------------------------------------------------------

#UTIL: Make a copy of the original faces
def solid_copy_group(lst_faces)
	if lst_faces.instance_of?(Sketchup::Group)
		lst_faces = lst_faces.entities.grep(Sketchup::Face)
	end	
	group = @entities.add_group
	gent = group.entities
	lst_faces.each { |face| G6.face_clone(face, gent, nil, @layer) }
	group
end

#UTIL: Determine on which side of a place is a face (1 or 2). 0 when on the section
def solid_face_which_side_of_plane(face, cut_plane)
	touch = false
	side = 0
	vec_plane = cut_plane[1]
	
	#Checking on which side of the plane
	face.vertices.each do |vx|
		pt = vx.position
		ptproj = pt.project_to_plane(cut_plane)
		ps = ptproj.vector_to(pt) % vec_plane
		if ps.abs > 0.000001
			side = (ps > 0) ? 1 : 2
			break
		end	
	end
	return [side, true, false] if side == 0
	
	#Checking if the face touches the plane
	touch = face.vertices.find { |vx| vx.position.on_plane?(cut_plane) }
	
	[side, touch]
end

#UTIL: Transfer faces from one group to another and erase the original ones
def solid_transfer_faces(faces, gent_to)
	return if faces.empty?
	hole_faces = faces.find_all { |f| f.loops.length > 1 }
	plain_faces = faces.find_all { |f| f.loops.length == 1 }
	faces = hole_faces + plain_faces
	
	faces.each do |face|
		G6.face_clone(face, gent_to, nil, @layer)
	end	
end

#UTIL: Eliminate tiny edges along the cut plane
def solid_manage_tiny_edges(gent, cut_plane)
	n = 0
	while true
		n += 1
		break if n > 1000
		tiny_edge = gent.grep(Sketchup::Edge).find { |e| e.length < 0.1 }
		break unless tiny_edge
		solid_delete_tiny_edge(tiny_edge, gent, cut_plane)
	end	
end

#SOLID: Eliminating a tiny edge if one of its extremity is on teh cut plane
def solid_delete_tiny_edge(edge, gent, cut_plane)
	vec_plane = cut_plane[1]
	pt1 = edge.start.position
	pt2 = edge.end.position
	lvx = lvec = line = nil
	if pt1.on_plane?(cut_plane)
		lvx = [edge.end]
		lvec = [pt2.vector_to(pt1)]
		line = [pt1, pt1.offset(vec_plane, 10)]
	elsif pt2.on_plane?(cut_plane)
		lvx = [edge.start]
		lvec = [pt1.vector_to(pt2)]
		line = [pt2, pt2.offset(vec_plane, 10)]
	end	
	return unless lvx
	
	gent.transform_by_vectors lvx, lvec
		
	#Creating a Fake line in a separate group to make the merge collapse the vertices 	
	g = gent.add_group
	e = g.entities.add_line *line
	g.entities.transform_by_vectors [e.end], [e.end.position.vector_to(e.start.position)]
	g.explode
end

#UTIL: Transfer_section_faces
def solid_transfer_section_faces(gent, cut_plane, vec_plane, reversed, diagonal)
	hed = {}
	ltouching_faces = gent.grep(Sketchup::Face)
	ltouching_faces.each do |face|
		face.edges.each do |e|
			pt1 = e.start.position
			pt2 = e.end.position
			if (pt1.on_plane?(cut_plane) && pt2.on_plane?(cut_plane))
				hed[e.entityID] = [e, pt1, pt2]
			end	
		end	
	end
		
	temp_g = gent.add_group
	temp_gent = temp_g.entities	
	
	#Transferring the edges
	led = []
	hed.each do |eid, a|
		edge, pt1, pt2 = a
		led.push temp_gent.add_line(pt1, pt2)
	end
	led.each { |e| e.find_faces }
	
	#Cleaning the faces, looking for holes to be created
	lerase = []
	temp_gent.grep(Sketchup::Face).each do |face|
		lerase.push face unless face.edges.find { |e| !G6.edge_coplanar?(e) }
		face.reverse! if face.normal % vec_plane > 0
	end
	temp_gent.erase_entities lerase
	
	#Removing the lonely edges
	temp_gent.erase_entities temp_gent.grep(Sketchup::Edge).find_all { |e| e.faces.length == 0 }
	
	#Exploding the temporary group
	lfaces = temp_g.explode.grep(Sketchup::Face)
	
	#Checking the orientation of the face and transfer of properties
	lfaces.each do |face|
		ledges = face.edges.find_all { |e| e.faces.length == 2 }
		#ledges = face.edges.sort { |e1, e2| e2.length <=> e1.length }
		ledges = ledges.sort { |e1, e2| e2.length <=> e1.length }
		edge0 = ledges.first
		break unless edge0
		f1, f2 = edge0.faces
		f = (f1 == face) ? f2 : f1
		face.reverse! if edge0.reversed_in?(face) == edge0.reversed_in?(f)
		G6.face_transfer_properties(f, face)
	end
end

#UTIL: Determine a reference face which is on or close to the concave edge
def solid_find_face_ref(lst_faces, pt1_ref, pt2_ref)
	face_ref = nil
	dmin = nil
	line_ref = [pt1_ref, pt2_ref]
	lst_faces.each do |face|
		ptinter = Geom.intersect_line_plane line_ref, face.plane
		next unless ptinter && [1, 2, 4].include?(face.classify_point(ptinter))
		d = [ptinter.distance(pt1_ref), ptinter.distance(pt2_ref)].min
		if !dmin || d < dmin
			dmin = d
			face_ref = face
		end	
	end
	face_ref
end

#UTIL: Split the faces on one side into connected groups of faces
def solid_face_grouping(hfaces, lst_touching_faces, pt1_ref, pt2_ref)
	lsgroupings = []
	hfaces_used = {}
	lst_faces = hfaces.values
	
	while true
		face0 = lst_faces.find { |f| !hfaces_used[f.entityID] }
		break unless face0
		hfaces_current = {}
		lsgroupings.push hfaces_current
		lsf = [face0]
		while lsf.length > 0
			face = lsf.shift
			face_id = face.entityID
			next if hfaces_used[face_id]
			hfaces_used[face.entityID] = true
			hfaces_current[face_id] = face
			faces = G6.face_neighbour_faces_by_edge(face).find_all { |f| fid = f.entityID ; hfaces[fid] && !hfaces_used[fid] }
			lsf.concat faces
		end	
	end
	
	#There is only one Grouping
	return [lsgroupings[0], []] if lsgroupings.length == 1

	#More than one grouping - Find a reference face
	face_ref = solid_find_face_ref(lst_touching_faces, pt1_ref, pt2_ref)
	return [hfaces, []] unless face_ref
	
	#Finding the grouping which contains the reference face
	face_ref_id = face_ref.entityID
	hgf_main = nil
	hgf_others = []
	lsgroupings.each do |hgf|
		if hgf[face_ref_id]
			hgf_main = hgf
		else
			hgf_others.push hgf
		end	
	end
	[hgf_main, hgf_others]
end

#UTIL: Create the cut plane, large enough to encompass the whole solid
def solid_prepare_cut_plane(pt0, vec_plane, diagonal)
	#Diagonal and rectangle points for the cut plane
	diag = diagonal
	diag2 = diagonal * 2
	vec1, vec2 = vec_plane.axes
	ptmid1 = pt0.offset vec2, diag
	pt1 = ptmid1.offset vec1, -diag
	pt2 = pt1.offset vec1, diag2
	pt3 = pt2.offset vec2, -diag2
	pt4 = pt3.offset vec1, -diag2
	
	#Creating the group with the cut plane
	cut_group = @entities.add_group
	cut_gent = cut_group.entities
	face = cut_gent.add_face [pt1, pt2, pt3, pt4]
	face.material = face.back_material = @mat_plane
	
	#Temporary plane for animation
	@group_plane.erase! if @group_plane
	@group_plane = solid_copy_group cut_group
	
	#Returning the cut plane group
	cut_group
end

#UTIL: Possibly split a group in several subgroups if faces are disjointed
def solid_disjointed_groups(top_group)
	top_gent = top_group.entities
	faces = top_gent.grep(Sketchup::Face)
	
	#No faces in the group. Erase it and return no group
	if faces.empty?
		top_group.erase!
		return []
	end	
	
	#Grouping faces which are connected
	lst_groupings = []
	hfaces_used = {}
	faces.each do |face|
		next if hfaces_used[face.entityID]
		con_faces = face.all_connected.grep(Sketchup::Face)
		con_faces.each { |f| hfaces_used[f.entityID] = true }
		lst_groupings.push con_faces
	end	
	
	#Only one grouping. Return the original group
	return [top_group] if lst_groupings.length == 1
	
	#Create groups for each grouping and delete the original one
	lsg = []
	lst_groupings.each do |faces|
		lsg.push solid_copy_group(faces)
	end
	top_group.erase!

	lsg
end

#--------------------------------------------------------------
# CONCAVE: Manage concave edges
#--------------------------------------------------------------

#CONCAVE: Build the list of concave edge information for the original solid
def concave_build_info(edges)
	t0 = Time.now
	#Inspecting the edges and identifying those which are concave
	nedges = 0
	angle_max_perp = 25.degrees
	angle_min_open = 140.degrees
	pi2 = 0.5 * Math::PI
	@lst_plane_info = []
	lst_perp = []
	lst_open = []
	lst_others = []
	edges.each do |edge|
		angle, face1, face2 = concave_edge_info(edge, @angle_concave_max)
		next unless angle
		nedges += 1
		pconcave = concave_create_info(edge, angle, face1, face2)
		if pconcave.angle_perp <= angle_max_perp
			lst_perp.push pconcave
		elsif pconcave.angle >= angle_min_open
			lst_open.push pconcave
		else	
			lst_others.push pconcave
		end	
	end	
	
	#No concave edges
	return nil if nedges == 0
	
	#Checking planes
	lst_perp.each do |pconcave|
		iplane1 = pconcave.plane1
		iplane2 = pconcave.plane2
		n1 = @lst_plane_info[iplane1][1]
		n2 = @lst_plane_info[iplane2][1]
		if n1 == n2
			choice = (pconcave.face1.area > pconcave.face2.area) ? 1 : 2
		elsif n1 > n2
			choice = 1
		else
			choice = 2
		end	
		if choice == 1
			pconcave.iplane_max = iplane1
			pconcave.nplane_max = n1
			pconcave.plane_area = pconcave.face1.area
			pconcave.best_plane = [pconcave.lptref, pconcave.normal1]
		else 	
			pconcave.iplane_max = iplane2
			pconcave.nplane_max = n2
			pconcave.plane_area = pconcave.face2.area
			pconcave.best_plane = [pconcave.lptref, pconcave.normal2]
		end	
	end
	
	#Handling of edges whose angle is close to 90 degrees
	lst_perp.sort! { |pcv1, pcv2| concave_sorting_perp(pcv1, pcv2) }
	#pcv = lst_perp[0]
	#pcv.edge.material = 'gold' if pcv

	#Handling of open edges
	lst_open.sort! { |pcv1, pcv2| concave_sorting(pcv1, pcv2) }
	#pcv = lst_open[0]
	#pcv.edge.material = 'orange' if pcv

	#Handling of open edges
	lst_others.sort! { |pcv1, pcv2| concave_sorting(pcv1, pcv2) }
	
	#Using Guide edges if any
	unless @param_box_style
		lst_perp.each do |pconcave|
			concave_use_guiding_edge(pconcave)
		end
	end
	
	#Building the final list of concave edge information	
	@lst_concave_info = lst_perp + lst_others + lst_open
	
	#Ranking the concave edges and refining their best plane 
	@lst_concave_info.each_with_index do |pconcave, i| 
		pconcave.rank = i
	end	

	#Returning status
	nedges
end

#CONCAVE: Register the planes at concave edges
def concave_register_plane(plane0)
	@lst_plane_info.each_with_index do |plane_info, i|
		plane, nb = plane_info
		if concave_plane_equal?(plane0, plane)
			plane_info[1] += 1
			return i
		end	
	end	
	@lst_plane_info.push [plane0, 1]
	@lst_plane_info.length-1
end

#CONCAVE: Check if two planes are the same
def concave_plane_equal?(plane1, plane2)
	pt1, normal1 = plane1
	pt2, normal2 = plane2
	return false unless normal1.parallel?(normal2)
	(pt2.on_plane?(plane1))
end

#CONCAVE: Check if there is a guiding edge
def concave_use_guiding_edge(pconcave)
	edge0 = pconcave.edge
	face1, face2 = edge0.faces
	vx_edges = edge0.start.edges + edge0.end.edges
	led = vx_edges.find_all { |e| e != edge0 && !e.used_by?(face1) && !e.used_by?(face2) }
	return if led.empty?
	
	vec_bissec = pconcave.vec_bissec
	ls = []
	led.each do |ed|
		vec = G6.vector_edge(ed)
		ps = (vec.normalize % vec_bissec.normalize).abs
		if ps > 0.85
			normal = vec * pconcave.vec_dir
			ls.push [ps, normal]
		end	
	end
	return if ls.empty?
	
	ls.sort! { |a, b| b.first <=> a.first }
	normal = ls[0][1]
	pconcave.best_plane = [pconcave.lptref, normal]
end

#CONCAVE: Compute the best cut plane based on the list of concave edges in the group
#         Return nil if the group is convex
def concave_best_cut_plane(group)
	gent = group.entities
	concave_edges = gent.grep(Sketchup::Edge).find_all { |e| concave_edge_info(e, @angle_concave_max) }
	return nil if concave_edges.empty?

	#Retrieving the best corresponding concave info structure
	lst_pconcave = concave_edges.collect { |edge| concave_edge_check_info(edge) }
	lst_pconcave = lst_pconcave.find_all { |pconcave| pconcave != nil }
	return nil if lst_pconcave.empty?
	
	#Sorting the edges
	lst_pconcave.sort! { |pcv1, pcv2| pcv1.rank <=> pcv2.rank }
	
	#Returning the best plane
	lst_pconcave.first.best_plane
end

#CONCAVE: Retrieve the parent concave edge (edge collinear)
def concave_edge_check_info(edge)
	pt1 = edge.start.position
	pt2 = edge.end.position
	@lst_concave_info.each do |pconcave|
		line = [pconcave.lptref[0], pconcave.vec_dir]
		return pconcave if pt1.on_line?(line) && pt2.on_line?(line)
	end
	nil
end

#CONCAVE: Create a concave edge info structure
def concave_create_info(edge, angle, face1, face2)
	#Creating the structure
	pconcave = ConcaveEdgeInfo.new
	pconcave.edge = edge
	pconcave.face1 = face1
	pconcave.face2 = face2
	pt1 = edge.start.position
	pt2 = edge.end.position
	
	#Characteristics of edge
	pconcave.vec_dir = G6.vector_edge(edge)
	pconcave.lptref = [pt1, pt2]
	pconcave.len = edge.length 
	pconcave.normal1 = normal1 = face1.normal
	pconcave.normal2 = normal2 = face2.normal
	pconcave.plane1 = concave_register_plane([pt1, normal1])
	pconcave.plane2 = concave_register_plane([pt1, normal2])
	pconcave.vec_bissec = normal1 + normal2
	pconcave.vec_bissec_normal = normal1 - normal2
	pconcave.angle = angle
	pconcave.angle_perp = ((Math::PI * 0.5) - angle).abs
	pconcave.best_plane = [pconcave.lptref, pconcave.vec_bissec_normal]
	
	pconcave
end

#CONCAVE: Compute the best cut plane from the concave edges
def concave_connected_edges
	hsh_concave_edges = {}
	@lst_concave_edge_info.each do |pconcave| 
		hsh_concave_edges[pconcave.edge.entityID] = pconcave
	end	
		
	#Analysing the concave edges	
	@lst_concave_edge_info.each do |pconcave|
		edge = pconcave.edge
		
		#Checking if the edge is connected to another concave edge
		nb = 0
		edge.vertices.each do |vx|
			vx.edges.each do |e|
				next if e == edge
				pcv = hsh_concave_edges[e.entityID]
				nb += 1 if pcv && !ls.include?(pcv)
			end	
		end
		pconcave.nb_connected = nb	
	end
end

#CONCAVE: Method to sort concave edges (perp)
def concave_sorting_perp(pcv1, pcv2)
	#Priority to nb of connection
	nbc1 = pcv1.nb_connected
	nbc2 = pcv2.nb_connected
	return(nbc2 <=> nbc1) if nbc1 != nbc2

	#Priority to nb of plane
	nplane1 = pcv1.nplane_max
	nplane2 = pcv2.nplane_max
	return(nplane2 <=> nplane1) if nplane1 != nplane2

	#Priority to angle
	angle1 = pcv1.angle_perp
	angle2 = pcv2.angle_perp
	return(angle1 <=> angle2) if (angle1 - angle2).abs > 0.00001
	
	#Priority to longer length
	len1 = pcv1.len
	len2 = pcv2.len
	len2 <=> len1
end

#CONCAVE: Method to sort concave edges (others)
def concave_sorting(pcv1, pcv2)
	#Priority to angle
	angle1 = pcv1.angle
	angle2 = pcv2.angle
	return(angle1 <=> angle2) if (angle1 - angle2).abs > 0.00001
	
	#Priority to longer length
	len1 = pcv1.len
	len2 = pcv2.len
	len2 <=> len1
end

#CONCAVE: Check if an edge is concave. A minimum angle (in radians) can be given
def concave_edge_info(edge, angle_max=nil)
	faces = edge.faces
	nfaces = faces.length
	return nil if nfaces < 2
	
	#Computing normals to adjacent faces. Skipping coplanar faces
	ls = []
	for i in 0..nfaces-2
		face1 = faces[i]
		for j in i+1..nfaces-1
			face2 = faces[j]
			next if face1 == face2
			angle = concave_edge_by_two_faces(edge, face1, face2, angle_max)
			ls.push [angle, face1, face2] if angle
		end
	end
	return nil if ls.empty?
	
	#Sorting to get the smallest angle
	ls.sort { |a, b| a.first <=> b.first }
	
	ls[0]
end	

#CONCAVE: Check if an edge is concave. A minimum angle (in radians) can be given
def concave_edge_by_two_faces(edge, face1, face2, angle_max=nil)
	#Computing normals to adjacent faces. Skipping coplanar faces
	normal1 = face1.normal.normalize
	normal2 = face2.normal.normalize
	vec_perp = normal1 * normal2
	return nil unless vec_perp.valid?
	
	#Checking the orientation of edge along its faces. 
	#If both are equal, one of the face is wrongly oriented
	rev1 = edge.reversed_in?(face1)
	rev2 = edge.reversed_in?(face2)
	return nil if rev1 == rev2
	
	#Test on the orientation of edge vis-a-vis normals
	vec_ed = edge.start.position.vector_to edge.end.position
	vec_ed = vec_ed.reverse if rev1
	ps = vec_ed % vec_perp
	
	#Not concave
	return nil if ps >= 0
	
	#Calculating the angle and checking if it considered concave
	angle = normal1.angle_between normal2.reverse
	if angle_max
		return nil if angle >= angle_max
	end
	
	angle
end

#--------------------------------------------------------------
# FACE: Processing of Faces
#--------------------------------------------------------------

#FACE: Split a concave fax into concave decomposition
def face_convexify_all(lst_faces)
	status = 0
	hsh_faces = {}
	lst_faces.each { |f| hsh_faces[f.entityID] = f }
	lst_faces.each do |face|
		if face.loops.length > 1 || polygon_is_concave?(face.outer_loop.vertices.collect { |vx| vx.position }, face.normal)
			face_convexify(face, hsh_faces)
			status += 1
		end	
	end
	[hsh_faces.values, status]
end

#FACE: Split a concave fax into concave decomposition
def face_convexify(face, hsh_faces)
	#Triangulating the face
	ptriangles, pts_face = face_triangulation(face)
	normal = face.normal
	
	ptriangles.each do |ptri|
		#@entities.add_face ptri.lpt
	end	
	#return 1
	
	#Progressive merge of triangles to form convex polygons
	@hsh_ptri_used = {}
	lst_lipoly = []
	ptriangles.each do |ptri|
		next if @hsh_ptri_used[ptri.id]
		lipoly = face_extend_triangle(ptriangles, ptri, pts_face, normal)
		lst_lipoly.push lipoly unless lipoly.empty?
	end
	
	#Creating the split of faces
	lst_lipoly.each_with_index do |lipoly, ipoly|
		lpt = lipoly.collect { |i| pts_face[i] }
		f = @entities.add_face lpt
		f.reverse if f.normal % normal < 0
		hsh_faces[f.entityID] = f
	end
end

#FACE: Compute the triangulation of a face
def face_triangulation(face)
	#Transforming the face vertices to flat plane
	normal = face.normal
	tr_axe = Geom::Transformation.axes ORIGIN, *normal.axes
	tr_axe_inv = tr_axe.inverse
	pts_face = face.vertices.collect { |vx| vx.position }
	
	#Computing the Delaunay triangulation
	pts_flat = pts_face.collect { |pt| tr_axe_inv * pt }
	delau = Traductor::DelaunayTriangulation2D.new
	ltriangles = delau.calculate pts_flat
	
	#Creating the pseudo-triangles, eliminating those outside of the face
	ptriangles = []
	ltriangles.each do |litri|
		lpt = litri.collect { |i| pts_face[i] }
		center = G6.curve_barycenter(lpt)
		next unless [1, 2, 4].include?(face.classify_point(center))
		ptri = PseudoTriangles.new
		ptri.id = ptriangles.length
		ptriangles.push ptri
		ptri.lipt = litri
		ptri.lpt = litri.collect { |i| pts_face[i] }	
	end
	
	#Returning the list of triangles
	[ptriangles, pts_face]
end

#FACE: Extend a triangle recursively to its neighbours to form the biggest convex polygon
def face_extend_triangle(ptriangles, ptri0, pts_face, normal)
	#Initializing current polygon
	lipoly = []
	
	lptri = [ptri0]
	while lptri.length > 0
		#Current triangle
		ptri0 = lptri.shift
		next if @hsh_ptri_used[ptri0.id]
		
		#Merging the triangle with the current polygon
		new_lipoly = polygon_merge_with_triangle(lipoly, ptri0.lipt)
		pts = new_lipoly.collect { |i| pts_face[i] }
		
		#Skip if the new polygon is concave
		next if polygon_is_concave?(pts, normal)
		@hsh_ptri_used[ptri0.id] = true
		
		#New polygon is convex and valid
		lipoly = new_lipoly
		
		#Computing the next neighbours triangles
		lptri += ptriangles.find_all { |ptri| ptri != ptri0 && !@hsh_ptri_used[ptri.id] && (ptri.lipt & lipoly).length == 2 }
	end	
	
	lipoly
end

#--------------------------------------------------------------
# POLY: Utility methods on Polygons
#--------------------------------------------------------------

#POLY: Merge a polygon and a triangle. The method assumes the polygon is convex
def polygon_merge_with_triangle(lipoly, litri)
	#no polygon defined yet. Just return the triangle
	return litri.clone if lipoly.empty?
	
	#Common points to polygon and triangles
	i1, i2 = lipoly & litri
	
	#New vertex is the one in the triangle which is not common
	inew = litri.find { |i| i != i1 && i != i2 }
	
	#Inserting the new vertex in the polygon
	n = lipoly.length
	ipos1 = lipoly.rindex i1
	ipos2 = lipoly.rindex i2
	new_lipoly = lipoly.clone
	if ipos1 == 0 && ipos2 == n-1
		new_lipoly.push inew
	else	
		new_lipoly[(ipos1+1).modulo(n), 0] = inew
	end	
	
	new_lipoly
end

#POLY: Determine if a polygon is concave
def polygon_is_concave?(pts, normal)
	return false if pts.length <= 3
	
	pi = Math::PI
	pi2 = 2 * pi
	
	livx = []
	n = pts.length - 1
	for i in 0..n
		j = (i == n) ? 0 : i+1
		pt0 = pts[i]
		vec1 = pt0.vector_to pts[i-1]
		vec2 = pt0.vector_to pts[j]
		vperp = vec1 * vec2
		next unless vperp.valid?
		angle = vec1.angle_between vec2
		angle = pi2 - angle if normal % vperp < 0
		return true if angle < pi
	end	
	false
end

end	#class SolidConvexifyCompact
	
end	#End Module Traductor
